#ifndef PHYSALGEBRA_HH
#define PHYSALGEBRA_HH

#include <iostream> // used in class Top
#include <functional> // used in class Selection

/*
 * This file contains the implementation of operators of a physical algebra.
 * When constructing a query plan, note that each plan starts with one tablescan
 * per relation and contains a single top operator.
 * Hence, the simplest query plan contains of a tablescan and a top operator.
 *
 * This file contains the following operators of a physical algebra:
 *   - Tablescan
 *   - Top
 *   - Selection <== ToDo: your task to implement the init, step and fin member functions
 *   - Projection
 *
 * the physical algebra implemented in this file has the following characteristics:
 *   - push-based
 *   - tuple-at-a-time processing
 *   - processes tuples in row format
 *   - assumes data is stored in row format
 */

template <typename Tconsumer>
class Tablescan {
  public:
    Tablescan (const relation_t& aRelation,
               Tconsumer&        aConsumerOp)
      : _relation(aRelation), _consumerOp(aConsumerOp)
      {}
  public:
    void run () {
      _consumerOp.init();
      tuple_t tuple;
      for (unsigned i = 0; i < _relation.size(); ++i) {
        tuple = _relation[i];
        _consumerOp.step(tuple);
      }
      _consumerOp.fin();
    }
  private:
    const relation_t& _relation;
    Tconsumer&  _consumerOp;
};


class Top {
  public:
    Top (bool aPrint, std::ostream& aOs = std::cout)
      : _print(aPrint), _os(aOs)
      {}
  public:
    void init() {
    }
    void step (const tuple_t& aTuple) {
      if (_print) {
        printTuple(aTuple, _os);
      }
    }
    void fin () {
    }
  private:
    bool          _print;
    std::ostream& _os;
};


template <typename Tconsumer>
class Selection {
  public:
    Selection (Tconsumer&                            aConsumerOp,
               std::function <bool (const tuple_t&)> aPredicate)
      : _consumerOp(aConsumerOp),
        _predicate(aPredicate)
      {}
  public:
    void init () {
      /*
       * Your code goes here.
       */
    }
    void step (const tuple_t& aTuple) {
      /*
       * Your code goes here.
       */
    }
    void fin () {
      /*
       * Your code goes here.
       */
    }
  public:
    Tconsumer& _consumerOp;
    // std::function encapsulates all callable types,
    // such as functions functors and lambdas.
    // http://www.cplusplus.com/reference/functional/function/function/
    std::function <bool (const tuple_t&)> _predicate;
};


template <typename Tconsumer>
class Projection {
  public:
    Projection (Tconsumer& aConsumerOp,
                std::vector<unsigned> aProjectionIndices)
      : _consumerOp(aConsumerOp),
        _projectionIndices(aProjectionIndices)
      {}
  public:
    void init () {
      _consumerOp.init();
    }
    void step (const tuple_t& aInputTuple) {
      tuple_t tuple;
      for (unsigned idx : _projectionIndices) {
        tuple.push_back( aInputTuple[idx] );
      }
      _consumerOp.step(tuple);
    }
    void fin () {
      _consumerOp.fin();
    }
  private:
    Tconsumer&            _consumerOp;
    std::vector<unsigned> _projectionIndices;
};

#endif
