#include "Relation.hh"

void
printTuple (const tuple_t& tuple, std::ostream& os) {
  for (unsigned i = 0; i < tuple.size(); ++i) {
    attrVal_t attrVal = tuple[i];
    printAttrVal(attrVal, os);
    os << " ";
  }
  os << std::endl;
}


void
printAttrVal (const attrVal_t& attrVal, std::ostream& os) {
  // simple way to print attrVal: pass callable (e.g. lambda function) to std::visit
  // std::visit executes callable for current type of value held.
  // Note: callable must be 
  auto printLambda = [&os] (const auto& val) {os << val;};
  std::visit (printLambda, attrVal);

  // complex way to print an attrVal
  /*
  switch(attrVal.index()) {
  case 0: // int, cf. attrVal_t
    os << std::get<int>(attrVal);
    break;
  case 1: // double, cf. attrVal_t
    os << std::get<double>(attrVal);
    break;
  case 2: // string, cf. attrVal_t
    os << std::get<std::string>(attrVal);
    break;
  default:
    throw std::invalid_argument("Unknown type");
  }
  */
}


Relation::Relation (const std::string& aRelationDescriptor, char delimiter) 
  : _data(), _schema(), _name(aRelationDescriptor) {
    readSchema("database/" + aRelationDescriptor + ".schema", delimiter);
    readData("database/" + aRelationDescriptor + ".data", delimiter);
}


const schema_t&
Relation::mygetSchema () {
  return _schema;
}


const relation_t&
Relation::getData () {
  return _data;
}


void
Relation::readSchema (const std::string& schemaDescriptor,
                      char               delimiter) {
  std::ifstream schemaFile (schemaDescriptor);
  std::string attributeSchema;
  while (std::getline(schemaFile, attributeSchema)) {
    std::string type = attributeSchema.substr(attributeSchema.find(delimiter)+1,
                                              attributeSchema.size());
    _schema.push_back(type);
  }
}


void
Relation::readData (const std::string& dataDescriptor,
                    const char         delimiter) {
  std::ifstream dataFile (dataDescriptor);
  std::string tupleString;
  while (std::getline(dataFile, tupleString)) { // read next tuple
    tuple_t tuple = readTuple(tupleString, delimiter);
    _data.push_back(tuple);
  }
}


tuple_t
Relation::readTuple (const std::string& tupleString,
                     const char         delimiter) {
  tuple_t tuple;
  std::istringstream attrValStream (tupleString);
  for (unsigned attrIndex = 0; attrIndex < _schema.size(); ++attrIndex) {
    attrVal_t attrVal = readAttrVal (attrValStream, attrIndex, delimiter);
    tuple.push_back(attrVal);
  }
  return tuple;
}


attrVal_t
Relation::readAttrVal (std::istringstream& attrValStream,
                      const unsigned      attrIndex,
                      const char          delimiter) {
  std::string attrValString;
  std::getline(attrValStream, attrValString, delimiter); // read next attrVal

  attrVal_t attrVal;
  if (_schema[attrIndex] == "int") {
    attrVal = std::stoi(attrValString);
  } else if(_schema[attrIndex] == "double") {
    attrVal = std::stod(attrValString);
  } else if(_schema[attrIndex] == "string") {
    attrVal = attrValString;
  } else {
    throw std::invalid_argument("Unknown type");
  }
  return attrVal;
}


void
Relation::print (std::ostream& os) {
  for (unsigned i = 0; i < _data.size(); ++i) {
    const tuple_t& tuple = _data[i];
    printTuple (tuple, os);
  }
}
