#ifndef PHYSALGEBRA_HH
#define PHYSALGEBRA_HH

#include <functional>
#include <vector>
#include <iostream>

#include "Relation.hh"
#include "../hashing/hashtable.hh"

/*
 * This file contains the implementation of operators of a physical algebra.
 * When constructing a query plan, note that each plan starts with one tablescan
 * per relation and contains a single top operator.
 * Hence, the simplest query plan contains of a tablescan and a top operator.
 *
 * This file contains the following operators of a physical algebra:
 *   - Tablescan
 *   - Top
 *   - Selection
 *   - Projection
 *   - Nested Loop Join (NestedLoopJoinOuter, NestedLoopJoinInner)
 *   - Hash Join (HashJoinBuild, HashJoinProbe)
 *
 * The physical algebra implemented in this file has the following characteristics:
 *   - push-based (memory flow, data flow and control flow)
 *   - tuple-at-a-time processing
 *   - processes tuples in row format
 *   - assumes data is stored in row format
 */


// Note: std::function encapsulates all callable types,
// such as functions functors and lambdas.
// http://www.cplusplus.com/reference/functional/function/function/
using selectionPred_t = std::function<bool (const tuple_t&)>;
using joinPred_t      = std::function<bool (const tuple_t&, const tuple_t&)>;
using hashfunction_t  = std::function<unsigned (const tuple_t&)>;


template <typename Tconsumer>
class Tablescan {
  public:
    Tablescan (const relation_t& aRelation,
               Tconsumer&        aConsumer,
               unsigned          aTupleSize=0)
      : _relation(aRelation),
        _consumer(aConsumer),
        _tuple()
      {
        _tuple.reserve(aTupleSize);
      }
  public:
    void run () {
      _consumer.init();
      for (unsigned i = 0; i < _relation.size(); ++i) {
        _tuple = _relation[i];
        _consumer.step(_tuple);
      }
      _consumer.fin();
    }
  private:
    const relation_t& _relation;
    Tconsumer&        _consumer;
    tuple_t           _tuple;
};


class Top {
  public:
    Top (bool aPrint, std::ostream& aOs = std::cout)
      : _print(aPrint), _os(aOs)
      {}
  public:
    void init() {
    }
    void step (tuple_t& aTuple) {
      if (_print) {
        printTuple(aTuple, _os);
      }
    }
    void fin () {
    }
  private:
    bool          _print;
    std::ostream& _os;
};


template <typename Tconsumer>
class Selection {
  public:
    Selection (Tconsumer&      aConsumer,
               selectionPred_t aPredicate)
      : _consumer(aConsumer),
        _predicate(aPredicate)
      {}
  public:
    void init () {
      _consumer.init();
    }
    void step (tuple_t& aTuple) {
      if (_predicate(aTuple)) {
        _consumer.step(aTuple);
      }
    }
    void fin () {
      _consumer.fin();
    }
  public:
    Tconsumer&      _consumer;
    selectionPred_t _predicate;
};


template <typename Tconsumer>
class Projection {
  public:
    Projection (Tconsumer&            aConsumer,
                std::vector<unsigned> aProjectionIndices)
      : _consumer(aConsumer),
        _projectionIndices(aProjectionIndices)
      {}
  public:
    void init () {
      _consumer.init();
    }
    void step (tuple_t& aInputTuple) {
      // Note: Simplifying assumption.
      // In a real system, no new tuple would be created.
      // Instead aInputTuple would be modified.
      tuple_t tuple;
      tuple.reserve(_projectionIndices.size());
      for (unsigned idx : _projectionIndices) {
        tuple.push_back( aInputTuple[idx] );
      }
      _consumer.step(tuple);
    }
    void fin () {
      _consumer.fin();
    }
  private:
    Tconsumer&            _consumer;
    std::vector<unsigned> _projectionIndices;
};


template <typename TscanInner, typename TNestedLoopJoinInner>
class NestedLoopJoinOuter {
  public:
    NestedLoopJoinOuter (const std::vector<TscanInner*>& aScanInnerPtrs,
                         TNestedLoopJoinInner&     aNljInner)
      : _scanInnerPtrs(aScanInnerPtrs),
        _nljInner(aNljInner)
      {}
    // second constructor just for convenience: can be used if inner table is result of only one scan
    NestedLoopJoinOuter (TscanInner&           aScanInner,
                         TNestedLoopJoinInner& aNljInner)
      : _scanInnerPtrs(1, &aScanInner),
        _nljInner(aNljInner)
      {}
  public:
    void init () {
    }
    void step (tuple_t& aTuple) {
      _nljInner._outerTuplePtr = &aTuple; // This class is friend of TNestedLoopJoinInner
      for (auto scan : _scanInnerPtrs) {
        scan->run();
      }
    }
    void fin () {
    }
  private:
    std::vector<TscanInner*> _scanInnerPtrs;
    TNestedLoopJoinInner&    _nljInner;
};


template <typename Tconsumer>
class NestedLoopJoinInner {
  public:
    NestedLoopJoinInner (Tconsumer&  aConsumer,
                         joinPred_t aPredicate)
      : _consumer(aConsumer),
      _joinPred(aPredicate),
      _outerTuplePtr(NULL)
      {}
  public:
    // Forbid compiler to default generate copy constructor and assignment operator
    NestedLoopJoinInner (const NestedLoopJoinInner&) = delete;
    NestedLoopJoinInner& operator= (const NestedLoopJoinInner&) = delete;
  public:
    void init () {
      _consumer.init();
    }
    void step (tuple_t& aInnerTuple) {
      if ( _joinPred(aInnerTuple, *_outerTuplePtr) ) {
        aInnerTuple.insert(aInnerTuple.begin(), _outerTuplePtr->begin(), _outerTuplePtr->end());
        _consumer.step(aInnerTuple);
      }
    }
    void fin () {
      _consumer.fin();
    }
  private:
    // Make NestedLoopJoinOuter a friend class
    // Allow class NestedLoopJoinOuter to access NestedLoopJoinInner's (this class) private/protected members.
    template<typename TscanInner, typename TNestedLoopJoinInner>
    friend class NestedLoopJoinOuter;
  private:
    Tconsumer& _consumer;
    joinPred_t _joinPred;
    tuple_t*   _outerTuplePtr; // points to a tuple produced by a NestedLoopJoinOuter instance
};


class HashJoinBuild {
  public:
    using HashTable_t = HashTable<tuple_t, hashfunction_t>;
  public:
    HashJoinBuild (unsigned              aNoBuckets,
                   const hashfunction_t& aBuildHashFunc)
      : _hashtable(aNoBuckets, aBuildHashFunc)
    {}
  public:
    void init () {
    }
    void step (tuple_t& aBuildTuple) {
      _hashtable.insert(aBuildTuple);
    }
    void fin () {
    }
  public:
    const HashTable_t& getHashTable () {
      return _hashtable;
    }
  private:
    HashTable_t _hashtable;
};


template <typename Tconsumer>
class HashJoinProbe {
  public:
    HashJoinProbe (Tconsumer&                        aConsumer,
                   const hashfunction_t&             aProbeHashFunc,
                   const HashJoinBuild::HashTable_t& aHashtable,
                   const joinPred_t&                 aJoinPred)
      : _consumer(aConsumer),
        _probeHashFunc(aProbeHashFunc),
        _hashtable(aHashtable),
        _joinPred(aJoinPred)
      {}
  public:
    void init () {
      _consumer.init();
    }
    void step (tuple_t aProbeTuple) {
      const unsigned probeTupleSize = aProbeTuple.size();
      const unsigned probeHashVal   = _probeHashFunc(aProbeTuple);

      for (auto bucketIter = _hashtable.getBucketIter(probeHashVal); bucketIter.isValid(); ++bucketIter) {
        const tuple_t& buildTuple = *bucketIter;
        if (_joinPred(buildTuple, aProbeTuple)) {
          // match found. concatenate buildTuple and aProbeTuple
          if (aProbeTuple.size() < probeTupleSize + buildTuple.size()) {
            aProbeTuple.resize(probeTupleSize + buildTuple.size());
          }
          for (unsigned i = 0; i < buildTuple.size(); ++i) {
            aProbeTuple[probeTupleSize + i] = buildTuple[i];
          }
          _consumer.step(aProbeTuple);
        }
      }
    }
    void fin () {
      _consumer.fin();
    }
  private:
    Tconsumer&                        _consumer;
    hashfunction_t                    _probeHashFunc;
    const HashJoinBuild::HashTable_t& _hashtable;
    joinPred_t                        _joinPred;
};

#endif
