#include <iostream>
#include <variant> // contains std::hash<std::variant>

#include "Relation.hh"
#include "PhysAlgebra.hh"


/*
 * This file contains the implementation of some queries over the test data set
 * and crimeInAtlanta2017 data set that contains crime records of Atlanta for
 * the year 2017. For details on the data see:
 * https://www.kaggle.com/priscillapun/crime-in-atlanta-2017
 * All datasets are stored in the subdirectory database. See Relation.hh for details.
 *
 * To implement a query we use the classes given in Relation.hh and physAlgebra.hh.
 * The operators given in physAlgebra.hh allow us to manually build a query plan
 * and to run it.
 *
 * Note on compilation: This implementation uses rather new C++17 features.
 * More specifically, it uses the std::variant class and its features.
 * Set the -std=c++17 compiler flag to compile.
 * It might be that not all compilers have implemented all C++17 features yet.
 * compile with:
 *   g++ -std=c++17 -Wall main.cc Relation.cc -o main.out
 * or use the given makefile.
 */

int main () {
  // load relations from disk to main memory
  Relation testRel ("test", ',');
  //testRel.print(std::cout);
  Relation atlantaRel ("crimeInAtlanta2017", ',');
  //atlantaRel.print(std::cout);
  Relation test2Rel ("test2", ',');
  //test2Rel.print(std::cout);

  {
    std::cout << "  === Executing query 0 ===" << std::endl;
    // SQL: SELECT * FROM test;

    bool print = true;
    Top            top (print, std::cout);
    Tablescan<Top> scan (testRel.getData(), top);

    scan.run();
  }


  {
    std::cout << "  === Executing query 1 ===" << std::endl;
    // Query 1: Find the names of all Persons named Olivia
    // SQL:     SELECT name FROM test WHERE name = 'Olivia';

    // predicate for Selection, implemented as lambda function
    auto nameEqConst = [] (const tuple_t& tuple) {return std::get<std::string>(tuple[1]) == "Olivia";};

    Top                                    top        (true, std::cout); 
    Projection <Top>                       projection (top, {1});
    Selection <Projection<Top>>            selection  (projection, nameEqConst);
    Tablescan <Selection<Projection<Top>>> scan       (testRel.getData(), selection);

    scan.run();
  }


  {
    std::cout << "  === Executing query 2 ===" << std::endl;
    // Query 2: List all crimes comitted in the neighborhood of Peachtree Hills
    // SQL: SELECT * FROM crimeInAtlanta2017 WHERE neighborhood = 'Peachtree Hills';

    // predicate for Selection, implemented as lambda function
    auto neighborhoodEqPeaHills = [] (const tuple_t& tuple) {return std::get<std::string>(tuple[19]) == "Peachtree Hills";};

    Top                        top       (true, std::cout);
    Selection <Top>            selection (top, neighborhoodEqPeaHills);
    Tablescan <Selection<Top>> scan      (atlantaRel.getData(), selection);

    scan.run();
  }


  {
    std::cout << "  === Executing query 3 ===" << std::endl;
    // Query 3: List the location and the maximum number of victims of 
    // crimes committed between longitude -84.36 and -84.35 and latitude between 33.73 and 33.74
    // SQL: SELECT location, MaxOfnum_victims FROM crimeInAtlanta2017
    //      WHERE x BETWEEN -84.36 AND -84.35 AND y BETWEEN 33.73 AND 33.74;

    // predicate for Selection, implemented as lambda function
    auto areaPred = [] (const tuple_t& tuple) {return std::get<double>(tuple[21]) >= -84.36 && std::get<double>(tuple[21]) <= -84.35 &&
                                                      std::get<double>(tuple[22]) >= 33.73  && std::get<double>(tuple[22]) <= 33.74;};

    Top                                    top        (true, std::cout);
    Projection <Top>                       projection (top, {10,14});
    Selection <Projection<Top>>            selection  (projection, areaPred);
    Tablescan <Selection<Projection<Top>>> scan       (atlantaRel.getData(), selection);

    scan.run();
  }


  {
    std::cout << "  === Executing query 4 ===" << std::endl;
    // Query 4: List the hometowns of each person (using the nested loop join).
    // SQL: SELECT * FROM test, test2 WHERE test == test2;

    auto joinPred = [] (const tuple_t& tuple1, const tuple_t tuple2) {return tuple1[0] == tuple2[0];};

    // template auto deduction: http://en.cppreference.com/w/cpp/language/template_argument_deduction
    Top                 top      (true, std::cout);
    NestedLoopJoinInner nljinner (top, joinPred);
    Tablescan           tblinner (test2Rel.getData(), nljinner);
    NestedLoopJoinOuter nljouter (tblinner, nljinner);
    Tablescan           tblouter (testRel.getData(), nljouter);

    tblouter.run();
  }


  {
    std::cout << "  === Executing query 5 ===" << std::endl;
    // Query 5: List the hometowns of each person (using the hash join operator).
    // SQL: SELECT * FROM test, test2 WHERE test == test2;

    unsigned noBucketsHashTable = 10;
    std::hash<attrVal_t> attrValHasher;
    auto hashFuncBuild = [&attrValHasher, noBucketsHashTable] (const tuple_t& tuple) {
      return attrValHasher(tuple[0]) % noBucketsHashTable;
    };
    // In general hashFuncProbe != hashFuncBuild, since build and probe tuple may have different layout,
    // i.e., the two hash function may access different indices of the tuples.
    auto hashFuncProbe = hashFuncBuild;
 
    auto joinPred = [] (const tuple_t& tuple1, const tuple_t& tuple2) {
      return tuple1[0] == tuple2[0];
    };

    HashJoinBuild hjbuild  (noBucketsHashTable, hashFuncBuild);
    Tablescan     tblbuild (test2Rel.getData(), hjbuild);

    Top           top      (true, std::cout);
    HashJoinProbe hjprobe  (top, hashFuncProbe, hjbuild.getHashTable(), joinPred);
    Tablescan     tblprobe (testRel.getData(), hjprobe);

    tblbuild.run();
    tblprobe.run();
  }
}
